/************************************************************************
 *
 * Constants and vars
 *
 ************************************************************************/
//~ nsIContentPolicy
//~ const REJECT_REQUEST            = Components.interfaces.nsIContentPolicy.REJECT_REQUEST;
//~ const REJECT_TYPE               = Components.interfaces.nsIContentPolicy.REJECT_TYPE;
//~ const REJECT_SERVER             = Components.interfaces.nsIContentPolicy.REJECT_SERVER;
//~ const REJECT_OTHER              = Components.interfaces.nsIContentPolicy.REJECT_OTHER;
const REJECT                    = Components.interfaces.nsIContentPolicy.REJECT_REQUEST;
const ACCEPT                    = Components.interfaces.nsIContentPolicy.ACCEPT;
const TYPE_IMAGE                = Components.interfaces.nsIContentPolicy.TYPE_IMAGE;
const TYPE_OBJECT               = Components.interfaces.nsIContentPolicy.TYPE_OBJECT;

//~ cache stuff
const CACHESERVICE              = Components.classes["@mozilla.org/network/cache-service;1"]
                                            .getService(Components.interfaces.nsICacheService);

var httpCacheSession            = CACHESERVICE.createSession("HTTP",
                                                             Components.interfaces.nsICache.STORE_ANYWHERE,
                                                             true);

httpCacheSession.doomEntriesIfExpired = false;

/**
    Note: During Gecko 1.8 development, nsIPrefBranchInternal was renamed to
    nsIPrefBranch2 [https://bugzilla.mozilla.org/show_bug.cgi?id=281414] and
    was frozen. nsIPrefBranchInternal name is still supported in Gecko 1.8,
    so this is what you should use in extensions that need to be compatible
    with Gecko 1.7 and Gecko 1.8 (Firefox 1.0/1.1).
    For newer extensions use nsIPrefBranch2.
**/

/************************************************************************
 *
 * Smthng...
 *
 ************************************************************************/

function nsImgLikeOpera() {
    Components.classes["@mozilla.org/observer-service;1"]
	          .getService(Components.interfaces.nsIObserverService)
	          .addObserver(this, "profile-after-change", false);
	
	//~ this.debug = true;
}

/**
 * JS XPCOM component registration
 */
nsImgLikeOpera.prototype = {
  
  log: function(msg) {
    if (!this.debug) return;
	msg = "[ILODebugMessage]: " + msg + "\n";
    this.console.logStringMessage(msg);
    dump(msg);
  },
  
  console:        Components.classes["@mozilla.org/consoleservice;1"]
	                        .getService(Components.interfaces.nsIConsoleService),
  
  prefBranch:     Components.classes["@mozilla.org/preferences-service;1"]
                            .getService(Components.interfaces.nsIPrefService)
						    .getBranch("imglikeopera."),
  
  windowMediator: Components.classes["@mozilla.org/appshell/window-mediator;1"]
                            .getService(Components.interfaces.nsIWindowMediator),
  
  settings: {
	expirationTime: 0,
	defaultPolicy: 3,
	tabsRefresh: false,
	policyPattrens: '',
	flashBlock: false,
	loadLinkedImages: 0,
	//~ delayLoad: false,
  },
  
  localizingDescription: function() {
    //~ localizing initial description here :o)
	//~ "Goofy-method", thanks Luana aka MatrixIsAllOver (extenzilla.it)
	
	var iloDescription   = Components.classes["@mozilla.org/supports-string;1"]
                                     .createInstance(Components.interfaces.nsISupportsString);
	
	iloDescription.data  = Components.classes["@mozilla.org/intl/stringbundle;1"]
                                     .createInstance(Components.interfaces.nsIStringBundleService)
									 .createBundle("chrome://imglikeopera/locale/imglikeopera.properties")
									 .GetStringFromName("initialDescription");
	
	var iloL18Prefs      = Components.classes["@mozilla.org/preferences-service;1"]
                                     .getService(Components.interfaces.nsIPrefService)
                                     .getBranch("extensions.imglikeopera@imfo.ru.")
									 .setComplexValue("description",
									                  Components.interfaces.nsISupportsString,
										              iloDescription);
  },
  
  /**
   * UserCSS
   */
  getUserCssFile: function() {
    var file = Components.classes["@mozilla.org/file/directory_service;1"]
                         .getService(Components.interfaces.nsIProperties)
                         .get("ProfD", Components.interfaces.nsIFile);

	file.append("extensions");
	file.append("imglikeopera@imfo.ru");
	file.append("chrome");
	file.append("imglikeopera");
	file.append("content");
	file.append("user.sheet.css");
	
	return file;
  },
  
  readUserCssRules: function(aCount) {
	aCount.value = 2;
	
	var fileContents = "";
	
	var is = Components.classes["@mozilla.org/network/file-input-stream;1"]
                       .createInstance( Components.interfaces.nsIFileInputStream);
	is.init(this.getUserCssFile(), 0x01, 0, is.CLOSE_ON_EOF);
	var sis = Components.classes["@mozilla.org/scriptableinputstream;1"]
                        .createInstance(Components.interfaces.nsIScriptableInputStream);
    sis.init(is);
	
    while(sis.available() > 0)
	  fileContents += sis.read(sis.available());
	
	is.close();
	sis.close();
    
    var match, rules=[];
    var reFFamily = new RegExp("family:\"(.*)\" !important", "i");
    var reFSize = new RegExp("size:(.*)px !important", "i");
    
    match = reFFamily.exec(fileContents);
    var fontfamily = (match == null) ? 'inherit' : match[1];
    
    match = reFSize.exec(fileContents);
    var fontsize = (match == null) ? 'inherit' : match[1];
    
    return [fontfamily, fontsize];
  },
  
  rewriteUserCss: function(fontfamily, fontsize) {
    fontfamily = fontfamily ? '"' + fontfamily + '"' : 'inherit';
    fontsize = fontsize ? fontsize + 'px' : 'inherit';
    
    var fileContents = "/* ImgLikeOpera bonus for userChrome.css */\n" +
					   "*[ilo-ph-fix] {\n" +
					   "-moz-force-broken-image-icon:1 !important;\n" +
                       "font-family:" + fontfamily + " !important;\n" +
                       "font-size:" + fontsize + " !important;\n" +
                       "font-weight:normal !important;\n" +
                       "line-height:140% !important;\n" +
                       "white-space:normal !important;\n" +
                       "}\n" +
                       "*[ilo-ph-fix='tofix'] {\n" +
					   "-moz-force-broken-image-icon:0 !important;\n" +
                       "border:1px inset !important;\n" +
                       "padding:1px 2px !important;\n" +
                       "white-space:pre !important;\n" +
                       "}\n" +
					   "embed.sIFR-flash[ilo-full-src] {\n" +
					   "-moz-binding: url('chrome://imglikeopera/content/imglikeopera.xml#sifr-alternate-restore') !important;\n" +
					   "}\n" +
					   "embed.sIFR-flash[ilo-ph-fix] {\n" +
					   "-moz-binding: url('chrome://imglikeopera/content/imglikeopera.xml#sifr-alternate') !important;\n" +
					   "}\n";
    
    var os = Components.classes["@mozilla.org/network/file-output-stream;1"]
                       .createInstance(Components.interfaces.nsIFileOutputStream);
	os.init(this.getUserCssFile(), 0x02 | 0x08 | 0x20, 0664, 0);
    var result = os.write(fileContents, fileContents.length);
	os.close();
	
	this.loadUserCss(); //~ reload
  },
  
  loadUserCss: function() {
    var sss = Components.classes["@mozilla.org/content/style-sheet-service;1"]
	                    .getService(Components.interfaces.nsIStyleSheetService);
    var ios = Components.classes["@mozilla.org/network/io-service;1"]
	                    .getService(Components.interfaces.nsIIOService);
    var u = ios.newURI("chrome://imglikeopera/content/user.sheet.css", null, null);
	
	//~ reload [but not worked...]
	if (sss.sheetRegistered(u, sss.USER_SHEET))
	  sss.unregisterSheet(u, sss.USER_SHEET);
	
	sss.loadAndRegisterSheet(u, sss.USER_SHEET);
  },
  
  loadSettings: function() {
	try {
	  this.settings.expirationTime   = this.prefBranch.getIntPref("expiration.time");
	  this.settings.defaultPolicy    = this.prefBranch.getIntPref("default.policy");
      this.settings.tabsRefresh      = this.prefBranch.getBoolPref("tabs.refresh");
	  this.settings.flashBlock       = this.prefBranch.getBoolPref("flash.block");
	  this.settings.loadLinkedImages = this.prefBranch.getIntPref("load.linked.images");
	  //~ this.settings.delayLoad        = this.prefBranch.getBoolPref("delay.load");
	  this.settings.policyPattrens   = this.createPatternsArray(this.prefBranch.getCharPref("policy.patterns"));
    } catch(ex) {}
  },
  
  /**
   * nsIObserver interface
   */
  observe: function(aSubject, aTopic, aData) {
    switch (aTopic) {
	  case "profile-after-change":
	    this.localizingDescription();
		break;
	  
	  case "app-startup":
	    //~ this.log("Hello! (ILOComponent from 'app-startup'-Observer)");
		var prefInternal = Components.classes["@mozilla.org/preferences-service;1"]
								     .getService(Components.interfaces.nsIPrefBranch2);
	    prefInternal.addObserver("imglikeopera", this, true);
		this.loadUserCss();
	  
	  default:
		this.loadSettings();
	}
  },
  
  shouldProcess: function(contentType, contentLocation, context, obj) {
	return ACCEPT;
  },
  
  /**
   * So... should load?
   */
  shouldLoad: function(contentType, contentLocation, context, obj) {
	if (contentLocation.scheme != "http" && contentLocation.scheme != "https")
	  return ACCEPT;
	
	/**
	 * TYPE_IMAGE: <img/>, <object type="image.."/>, cssBackgrounds...
	 * TYPE_OBJECT: flash... (and?..)
	 */
	if (contentType != TYPE_IMAGE && (contentType != TYPE_OBJECT || !this.settings.flashBlock))
	  return ACCEPT;
	
	var doc, fromCss = false;
	
	try { // <img/>
	  doc = obj.ownerDocument.defaultView.top.document;
	} catch(e) {
	  try { // cssBackground
	    doc = obj.QueryInterface(Components.interfaces.nsIDOMDocument);
		fromCss = true;
	  } catch(e) { // <object/>
	    try {
		  doc = obj.QueryInterface(Components.interfaces.nsIDOMNode).ownerDocument;
		  fromCss = false;
		} catch(e) {
		  //~ this.log("Catching document object... Ugggh... Something wrong.\r\nMaybe favicon, but who knows...");
		  return ACCEPT;
		}
	  }
	}
	
	if (!doc.iloDocPolicy) {
	  //~ var win, tabs, enumerator = this.windowMediator.getEnumerator("navigator:browser");
	  var tabs, enumerator = this.windowMediator.getEnumerator("navigator:browser");
	  while (enumerator.hasMoreElements()) {
	    tabs = enumerator.getNext().window.getBrowser().browsers;
		for (var i = tabs.length; --i > -1;) {
		  if (doc == tabs[i].contentDocument) {
		    doc.iloDocPolicy = tabs[i].iloTabPolicy;
			doc.iloUrlsBlock = doc.iloUrlsPass = "";
		  }
		}
	  }
	}
	
	if (!doc.iloDocPolicy) {
	  //~ this.log("Catching document object... Ugggh... Something wrong.");
	  return ACCEPT; //~ smthng wrng
	}
	
	var docPolicy = doc.iloDocPolicy;
	
	if (!fromCss) {
	  if (obj.getAttribute("ilo-ph-fix"))
	    return obj.getAttribute("ilo-ph-fix") == "mustLoad" ?
		                                         this.passObj(obj) :
												 this.blockObj(obj, contentLocation);
	  
	  //~ for all non-css objects:
	  obj.setAttribute("ilo-full-src", contentLocation.spec);
	}
	
	if (docPolicy == 3 && !fromCss && this.settings.loadLinkedImages && obj.parentNode.tagName == "A") {
	  //~ Load linked images in "Load cached images" mode
	  docPolicy = this.settings.loadLinkedImages == 1 ?
	                  1 : doc.location.host == obj.parentNode.host ?
					  1 : docPolicy;
	} else {
	  var re = new RegExp(contentLocation.spec.replace(/(\.|\||\+|\?|\$|\(|\)|\[|\])/g,"\\\$1")+" ");
	  if (re.test(doc.iloUrlsBlock))
	    return fromCss ? REJECT : this.blockObj(obj, contentLocation);
	  else if (re.test(doc.iloUrlsPass))
	    return fromCss ? ACCEPT : this.passObj(obj);
    }
	
	//~ check requested URL in USER filters
	docPolicy = this.getPolicyFromFilters(doc.location, contentLocation.spec) || docPolicy;
	
	//~ "Don't load images" => skip work with cache, return false.
	if (docPolicy == 4)
	  return fromCss ? REJECT : this.blockObj(obj, contentLocation);
	
	var expire = this.checkUrlInCache(contentLocation.spec, docPolicy);
	var passIt = true;
	
	switch (docPolicy) {
	  case 2:
	    //~ re = new RegExp(doc.location.host.replace(/^www\./i,""), "i");
		//~ passIt = re.test(contentLocation.host) ? true : false;
		passIt = doc.location.host == contentLocation.host ? true : false;
		break;
	  case 3:
	    passIt = expire ? false : true;
		break;
	}
	
	switch (passIt) {
	  case false:
	    doc.iloUrlsBlock += contentLocation.spec + " ";
		return fromCss ? REJECT : this.blockObj(obj, contentLocation);
		break;
	  case true:
	    doc.iloUrlsPass += contentLocation.spec + " ";
		return fromCss ? ACCEPT : this.passObj(obj);
		break;
	}
	
	//~ problems? => always return true
	return ACCEPT;
  },
  
  blockObj: function(obj, contentLocation) {
	
	//~ It'll be just "Submit Query" button
    if (obj.tagName.toLowerCase() == "input") return REJECT;
	
	if (obj.getAttribute("alt") == null) obj.setAttribute("alt", "[Image]");
	
	var oW = parseInt(obj.width || obj.style.width);
	var oH = parseInt(obj.height || obj.style.height);
	
	//~ set placeholder size
	if (!oW && !oH) { obj.setAttribute("ilo-ph-fix", "tofix"); return REJECT; }
	else if (oW && !oH) { obj.iloFixSH = true; obj.height = parseInt(oW / 0.668); }
	else if (!oW && oH) { obj.iloFixSW = true; obj.width = parseInt(oH * 0.668); }
	
	obj.setAttribute("ilo-ph-fix", "fixed");
	
	return REJECT;
  },
  
  passObj: function(obj) {
	obj.removeAttribute("ilo-ph-fix");
    if (obj.iloFixSH) { obj.iloFixSH = null; obj.removeAttribute("height"); }
    if (obj.iloFixSW) { obj.iloFixSW = null; obj.removeAttribute("width"); }
    return ACCEPT;
  },
  
  checkUrlInCache: function(url, docpolicy) {
	var expire = true;
	var timenow = parseInt((new Date().getTime())/1000-1);
	
	if (this.settings.expirationTime != 0) {
	  try {
	    var cacheEntryDescriptor = httpCacheSession.openCacheEntry(url,
		                               Components.interfaces.nsICacheEntryDescriptor,
									   false);
        var expirationTime = cacheEntryDescriptor.lastModified + this.settings.expirationTime;
		if (expirationTime < timenow) {
		  cacheEntryDescriptor.doom();
		} else if (cacheEntryDescriptor.expirationTime != expirationTime) {
		  cacheEntryDescriptor.setExpirationTime(expirationTime);
		  cacheEntryDescriptor.markValid();
		  expire = false;
		}
		cacheEntryDescriptor.close();
		/************************************************************************
		 *
		 * "Я не знаю кому и зачем это нужно" (c) Вертинский
		 *
		 * Тут такая тема... Если повторно не обратиться сейчас к кэшу, то
		 * "setExpirationTime" (см. выше) не сработает, что всплывает при
		 * перезагрузке страницы.
		 *
		 * 2Do: Надо бы разобраться и поковырять.
		 * Позже: во всяком случае, так было в FF 0.9.3
		 * Ещё позже: ff1.0 -- ага, так и есть
		 *
		 ************************************************************************/
		 try {
		   cacheEntryDescriptor = httpCacheSession.openCacheEntry(url,
		                              Components.interfaces.nsICache.ACCESS_READ,
									  false);
           if (cacheEntryDescriptor)
		     expire = false;
		 } catch(ex) {}
            
      } catch(ex) {}
    
	} else if (docpolicy == 3) {
	  try {
	    var cacheEntryDescriptor = httpCacheSession.openCacheEntry(url,
		                               Components.interfaces.nsICache.ACCESS_READ,
									   false);
        if (cacheEntryDescriptor.expirationTime > timenow)
		  expire = false;
      } catch(ex) {}
    }
	
	return expire;
  },
  
  getPolicyFromFilters: function(docLocation, url) {
	var policy, patts = this.settings.policyPattrens;
	for (var i=0; i<patts.length; i++) {
	  if (patts[i][0].test(docLocation) || patts[i][0].test(url)) {
	    policy = patts[i][1];
		break;
	  }
    }
    return policy;
  },
  
  createPatternsArray: function(str) {
    var patt, pattern = [], tmp = str.split("  ");
    var tmpCount = tmp.length-1;
    for (var i=0; i<tmpCount; i++) {
	  patt = tmp[i].split(" ");
	  patt[1] = patt[1][0]-0;
	  if (patt.length == 2 && patt[1]>0 && patt[1]<5)
	    pattern.push([this.txt2RegExp(patt[0]), patt[1]]);
	}
	return pattern;
  },
  
  txt2RegExp: function(txt) {
    //~ not simple; return power
	if (txt[0] == "/" && txt[2]) {
	  var mod = "", txtEnd = txt.substr(txt.length-2, txt.length);
	  
	  if (txtEnd == "/i" || txtEnd == "/I") { mod = "i"; var res = txt.substr(1, txt.length-3); }
	  else if (txtEnd[1] == "/") { var res = txt.substr(1, txt.length-2); }
	  
	  try { var reg = new RegExp(res, mod); } catch(e){}
	  
	  if (reg) return reg;
	}
	
	//~ else: simple;
	s = new String(txt);
	var sCount = s.length;
	var res = new String("^");
	
	for (var i=0 ; i<sCount; i++) {
	  switch(s[i]) {
	    case '*' :
		  res += '.*';
		  break;
		
		case '.' :
		case '?' :
		case '^' :
		case '$' :
		case '+' :
		case '{' :
		case '[' :
		case '|' :
		case '(' :
		case ')' :
		case ']' :
		  res += '\\' + s[i];
		  break;
		
		case '\\' :
		  res += '\\\\';
		  break;
		
		default :
		  res += s[i];
		  break;
	  }
    }
	
	return new RegExp(res, 'i');
  },
  
  QueryInterface: function(iid) {
    if (!iid.equals(Components.interfaces.nsIImgLikeOpera) &&
	    !iid.equals(Components.interfaces.nsISupports) &&
        !iid.equals(Components.interfaces.nsISupportsWeakReference) &&
        !iid.equals(Components.interfaces.nsIContentPolicy) &&
        !iid.equals(Components.interfaces.nsIObserver))
    {
	  dump("nsImgLikeOpera: QI unknown interface: " + iid + "\n");
      throw Components.results.NS_ERROR_NO_INTERFACE;
	}
	return this;
  },
  
  /**
   * Component's module implementation
   */
  module: {
    // registerSelf: Register this component.
    registerSelf: function (compMgr, fileSpec, location, type) {
      var compReg = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
      compReg.registerFactoryLocation(this.cid,
                                      "nsImgLikeOpera JS component",
                                      this.contractId,
                                      fileSpec, location, type );
	  var catman = Components.classes["@mozilla.org/categorymanager;1"]
	                         .getService(Components.interfaces.nsICategoryManager);
	  catman.addCategoryEntry("app-startup", "nsImgLikeOpera", this.contractId, true, true);
	  catman.addCategoryEntry("content-policy", this.contractId, this.contractId,true,true);
	},
	
	unregisterSelf: function(compMgr, fileSpec, location) {
	  compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
	  compMgr.unregisterFactoryLocation(this.cid, fileSpec);
      var catman = Components.classes["@mozilla.org/categorymanager;1"]
                             .getService(Components.interfaces.nsICategoryManager);
      catman.deleteCategoryEntry("app-startup", this.contractId, true);
	  catman.deleteCategoryEntry("content-policy", this.contractId, true);
	},
	
	// getClassObject: Return this component's factory object.
	getClassObject: function (compMgr, cid, iid) {
	  if (!cid.equals(this.cid))
	    throw Components.results.NS_ERROR_NO_INTERFACE;
	  if (!iid.equals(Components.interfaces.nsIFactory))
	    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
	  return this.factory;
	},
	
	/* CID for this class */
	// extensions.roachfiend.com/cgi-bin/guid.pl
	cid: Components.ID("{9aa46f4f-4dc7-4c06-97af-5035170633fe}"),
	
	/* Contract ID for this class */
	contractId: "@mozilla.org/imglikeopera;1",
	
	/* factory object */
    factory: {
	  // createInstance: Return a new nsImgLikeOpera object.
      createInstance: function (outer, iid) {
        if (outer != null)
		  throw Components.results.NS_ERROR_NO_AGGREGATION;
		return new nsImgLikeOpera();
      }
	},
	
	canUnload: function(compMgr) {
	  return true;
	}
  }
}

// NSGetModule: Return the nsIModule object.
function NSGetModule(compMgr, fileSpec) {
    return nsImgLikeOpera.prototype.module;
}
